/*
 * REX-PCI20, GPIB Driver - Ver 1.0
 *
 * Copyright (C) 2007-2008 RatocSystems, Inc.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation, version 2.
 *
 * Based on dummy PCMCIA client driver
 * 
 * Changelog:
 *  2007/12/17  start coding (kernel 2.6.18-1.2798.fc6)
 *  2008/07/16  modified.
 *
 * All rights reserved, Dual BSD/GPL Licensed.
 *
 **/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/version.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <linux/fs.h>		/* for struct file_operations */
#include <linux/wait.h>
#include <linux/interrupt.h>
#include <linux/pci.h>

#include "rexpci20.h"
#include "def9914.h"

///#define PCI_DEBUG

#define MY_MAJOR 253
#define MY_MINOR 0
#define MY_TIMEOUT	(2*HZ)			/* 2 seconds */
#define CM_MAX_DEV 1 


/* Structure to hold all of our device specific stuff */
struct rs_gpib {
	int	open;
	struct pci_dev	*p_dev;
	wait_queue_head_t waitq;
	int	waitq_complete;
	struct semaphore lock;          /* general race avoidance */
};

static struct pci_dev *dev_table[CM_MAX_DEV];

/*====================================================================*/

/* Module parameters */

static int major;
static struct class *gpib_classgpib_class;

/*
   All the PCMCIA modules use PCMCIA_DEBUG to control debugging.  If
   you do not define PCMCIA_DEBUG at all, all the debug code will be
   left out.  If you compile with PCMCIA_DEBUG=0, the debug code will
   be present but disabled -- but it can then be enabled for specific
   modules at load time with a 'pc_debug=#' option to insmod.
*/

#ifdef PCI_DEBUG
//INT_MODULE_PARM(pc_debug, PCI_DEBUG);
//#define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG args)
//#define DEBUG(n, args...) printk(KERN_DEBUG args)
static char *version = "ver 1.0";
#else
#define DEBUG(n, args...)
#endif


//Define our PCI vendor/device ID
#define PCI_VENDOR_ID_20		0x1195
#define PCI_DEVICE_ID_20		0x2001

static struct pci_device_id rs_gpib_ids[] __devinitdata = {
	{ PCI_VENDOR_ID_20, PCI_DEVICE_ID_20, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{ 0, }
};
MODULE_DEVICE_TABLE(pci, rs_gpib_ids);


/* Interrupt service routine entry point.
 * 	
 * Arguments:
 * 
 * irq     interrupt number that caused interrupt
 * dev_id  device ID supplied during interrupt registration
 * regs    interrupted processor context
 */
static void rexpci20_isr(int irq, void *dev_id, struct pt_regs * regs)
{
	struct rs_gpib *dev = (struct rs_gpib *)dev_id;
	int iobase = pci_resource_start(dev->p_dev, 0);

#ifdef PCI_DEBUG
	printk(KERN_INFO "------ rexpci20_isr irq = 0x%x\n", irq);
	//printk(KERN_INFO "------ rexpci20_isr isr0 = 0x%x isr1 = 0x%x\n", inb(iobase+0), inb(iobase+1));
#endif

	// change_071225 TMS9914 interrupt disable
	outb(0, iobase+IMR0);
	outb(0, iobase+IMR1);
	outb(piimr2, iobase+AUXCR);
	outb(0, iobase+IMR2);
	outb(clrpi, iobase+AUXCR);
	dev->waitq_complete = 1;
	wake_up_interruptible(&dev->waitq);
	//free_irq(irq, dev);
}


ssize_t gpib_read(struct file *file, char *buffer, size_t count, loff_t *ppos)
{
#ifdef PCI_DEBUG
	printk(KERN_INFO "gpib_read count = 0x%x\n", count);
#endif
	return 0;
}

ssize_t gpib_write(struct file *file, const char *user_buffer, size_t count, loff_t *ppos)
{
#ifdef PCI_DEBUG
	printk(KERN_INFO "gpib_write count = 0x%x\n", count);
#endif
	return 0;
}

int gpib_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct rs_gpib *dev = (struct rs_gpib *)file->private_data;
	int iobase = pci_resource_start(dev->p_dev, 0);
	CARD_RESOURCE		card;
	IOPORT_UCHAR		bIOPort;
	IOPORT_USHORT		wIOPort;
	//void __user *data;
	int retval = 0;

	down(&(dev->lock));
	/* Sanity check to make sure rio is connected, powered, etc */
    if ( dev == NULL)
	{
		retval = -ENODEV;
		goto err_out;
	}

	switch (cmd) {
	case IOCTL_GP_GET_RESOURCE:
		card.IOBaseLow = iobase;
		card.IOBaseHigh = pci_resource_end(dev->p_dev, 0);
		card.IOLength = pci_resource_len(dev->p_dev, 0);
		card.IrqLevel = dev->p_dev->irq;
		if (copy_to_user((int *)arg, &card, sizeof(CARD_RESOURCE))) {
			retval = -EFAULT;
			goto err_out;
		}
		break;
	case IOCTL_GP_READ_PORT_UCHAR:
		if (copy_from_user(&bIOPort, (int *)arg, sizeof(IOPORT_UCHAR))) {
			retval = -EFAULT;
			goto err_out;
		}
		bIOPort.Value = inb(iobase+bIOPort.Offset);
		if (copy_to_user((int *)arg, &bIOPort, sizeof(IOPORT_UCHAR))) {
			retval = -EFAULT;
			goto err_out;
		}
		break;
	case IOCTL_GP_READ_PORT_USHORT:
		if (copy_from_user(&wIOPort, (int *)arg, sizeof(IOPORT_USHORT))) {
			retval = -EFAULT;
			goto err_out;
		}
		wIOPort.Value = inw(iobase+wIOPort.Offset);
		if (copy_to_user((int *)arg, &wIOPort, sizeof(IOPORT_USHORT))) {
			retval = -EFAULT;
			goto err_out;
		}
		break;
	case IOCTL_GP_WRITE_PORT_UCHAR:
		if (copy_from_user(&bIOPort, (int *)arg, sizeof(IOPORT_UCHAR))) {
			retval = -EFAULT;
			goto err_out;
		}
		outb(bIOPort.Value, iobase+bIOPort.Offset);
		break;
	case IOCTL_GP_WRITE_PORT_USHORT:
		if (copy_from_user(&wIOPort, (int *)arg, sizeof(IOPORT_USHORT))) {
			retval = -EFAULT;
			goto err_out;
		}
		outw(wIOPort.Value, iobase+wIOPort.Offset);
		break;
	case IOCTL_GP_START_INTERRUPT:
#ifdef PCI_DEBUG
		printk(KERN_INFO "gpib_ioctl: IOCTL_GP_START_INTERRUPT\n");
		printk(KERN_INFO "gpib_ioctl INTERRUPT...->sleep_on\n");
#endif
		up(&(dev->lock));
		//request_irq(dev->p_dev->irq, (void *)rexpci20_isr, SA_INTERRUPT|SA_SHIRQ, "rexpci20", dev);
		//++ 080716 - TMS9914 interrupt enable
		outb(0x00, iobase+IMR0);
		outb(0x02, iobase+IMR1);
		outb(piimr2, iobase+AUXCR);
		outb(0xC0, iobase+IMR2);
		outb(clrpi, iobase+AUXCR);
		//--
		dev->waitq_complete = 0;
		wait_event_interruptible(dev->waitq, (dev->waitq_complete) );
		down(&(dev->lock));
#ifdef PCI_DEBUG
		printk(KERN_INFO "gpib_ioctl INTERRUPT...sleep_on->\n");
#endif
		break;
	case IOCTL_GP_STOP_INTERRUPT:
#ifdef PCI_DEBUG
		printk(KERN_INFO "gpib_ioctl: IOCTL_GP_STOP_INTERRUPT\n");
#endif
		//++ 080716 - TMS9914 interrupt disable
		outb(0, iobase+IMR0);
		outb(0, iobase+IMR1);
		outb(piimr2, iobase+AUXCR);
		outb(0, iobase+IMR2);
		outb(clrpi, iobase+AUXCR);
		//--
		dev->waitq_complete = 1;
		wake_up_interruptible(&dev->waitq);
		break;
	default:
		retval = -ENOIOCTLCMD;
		break;
	}

err_out:
	up(&(dev->lock));
	return retval;
}

int gpib_open(struct inode *inode, struct file *file)
{
	struct rs_gpib *dev;
	struct pci_dev *link;
#if(LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
	int minor = MINOR(inode->i_rdev);
#else
	int minor = iminor(inode);
#endif

#ifdef PCI_DEBUG
		printk(KERN_INFO "gpib_open\n");
#endif

	if (minor >= CM_MAX_DEV)
		return -ENODEV;

	link = dev_table[minor];
	if (link == NULL)
		return -ENODEV;

	dev = link->sysdata;
	if (dev->open)
		return -EBUSY;
	file->private_data = dev;

	init_waitqueue_head(&dev->waitq);
	init_MUTEX(&(dev->lock));

	if (file->f_flags & O_NONBLOCK) {
#ifdef PCI_DEBUG
		printk(KERN_INFO "filep->f_flags O_NONBLOCK set\n");
#endif
		return -EAGAIN;
	}

	dev->open = 1;

#ifdef PCI_DEBUG
	printk(KERN_INFO "gpib_open version = %s minor = %d\n", version, minor );
	printk(KERN_INFO "io start = 0x%04x end = 0x%04x\n", 
		(int)pci_resource_start(link, 0), (int)pci_resource_end(link, 0));
	printk(KERN_INFO "io len = 0x%x, irq = %d\n", 
		(int)pci_resource_len(link, 0), link->irq);
#endif

	return 0;
}

int gpib_release(struct inode *inode, struct file *file)
{
	struct rs_gpib *dev;
	struct pci_dev *link;
#if(LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
	int minor = MINOR(inode->i_rdev);
#else
	int minor = iminor(inode);
#endif

#ifdef PCI_DEBUG
	printk(KERN_INFO "gpib_release subminor = %d\n", minor);
#endif
	dev = (struct rs_gpib *)file->private_data;

	if (minor >= CM_MAX_DEV)
		return -ENODEV;

	link = dev_table[minor];
	if (link == NULL)
		return -ENODEV;

	if(dev->open == 1){
		dev->open = 0;
	}

	return 0;
}

static struct file_operations rexpci20_fops = {
	.owner =	THIS_MODULE,
	.read =		gpib_read,
	.write =	gpib_write,
	.ioctl =	gpib_ioctl,
	.open =		gpib_open,
	.release =	gpib_release,
};


/*======================================================================

    rexpci20_probe() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

======================================================================*/
static int rexpci20_probe(struct pci_dev *link, const struct pci_device_id *id)
{
	struct rs_gpib *dev;
	int i;

#ifdef PCI_DEBUG
	printk(KERN_INFO "rexpci20_probe()\n");
#endif

	for (i = 0; i < CM_MAX_DEV; i++) {
		if (dev_table[i] == NULL)
			break;
	}

	if (i == CM_MAX_DEV)
		return -ENODEV;

    /* Create new parport device */
	dev = (struct rs_gpib *)kmalloc(sizeof(struct rs_gpib), GFP_KERNEL);
	if (dev == NULL)
		return -ENOMEM;
	memset(dev, 0, sizeof(*dev));
	link->sysdata = dev;
	dev->p_dev = link;

	dev_table[i] = link;

	pci_enable_device(link);
	request_irq(link->irq, (void *)rexpci20_isr, SA_INTERRUPT|SA_SHIRQ, "rexpci20", dev);

	/* register this driver*/
	major = register_chrdev(0, "gpib", &rexpci20_fops);
	if (major < 0) {
#ifdef PCI_DEBUG
		printk(KERN_INFO "register_chrdev failed. Error number %d", major);
#endif
	}
#ifdef PCI_DEBUG
	printk(KERN_INFO "major = %d\n", major);
#endif

#if 0//(LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
	// mknod -m 666 /dev/gpib0 c MY_MAJOR 0
	gpib_class = class_create(THIS_MODULE, "rexpci20_gpib");
	if (!gpib_class)
		return -1;
	class_device_create(gpib_class, NULL, MKDEV(major, i), NULL,
			    "gpib%d", i);
#endif


    return 0;
} /* rs_gpib_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/
static void rexpci20_remove(struct pci_dev *link)
{
	struct rs_gpib *dev = link->sysdata;
	int i;

	/* find device */
	for (i = 0; i < CM_MAX_DEV; i++) {
		if (dev_table[i] == link)
			break;
	}
	if (i == CM_MAX_DEV)
		return;

#ifdef PCI_DEBUG
	printk(KERN_INFO "rexpci20_remove(0x%p) i = %d\n", link, i);
#endif

	free_irq(link->irq, dev);
	pci_disable_device(link);

	dev_table[i] = NULL;
	kfree(dev);

#if 0//(LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
	// rm /dev/gpib0
	class_device_destroy(gpib_class, MKDEV(major, i));
	class_destroy(gpib_class);
#endif

	unregister_chrdev(major, "gpib");
    
} /* rs_gpib_detach */


static struct pci_driver rexpci20_driver = {
	.name			= "rexpci20",
	.probe		= rexpci20_probe,
	.remove		= rexpci20_remove,
	.id_table	= rs_gpib_ids,
};


static int __init rexpci20_init_module(void)
{
	int rc;

#ifdef PCI_DEBUG
	printk(KERN_INFO "rexpci20_init_module: loading...\n");
#endif

	rc = pci_register_driver(&rexpci20_driver);
	if (rc < 0) {
		return rc;
	}

	return 0;
}

static void __exit rexpci20_exit_module(void)
{
#ifdef PCI_DEBUG
	printk(KERN_INFO "rexpci20_exit_module: unloading...\n");
#endif

	/* unregister this driver */
	pci_unregister_driver(&rexpci20_driver);
}

module_init (rexpci20_init_module);
module_exit (rexpci20_exit_module);
MODULE_LICENSE("Dual BSD/GPL");

